/*
 * InCode - [ compile.c ]
 * (c) 2002, J. Weeks
 * jweeks@subdimension.com
 *
 * This part of the compiler will actually take the tokens, in modified
 * order, and produce assembly language output.
 *
 * NOTE:
 *   At this point in time 2*(1+1) compiles better then (1+1)*2 and has to
 *   do with pushing and poping to the stack, and the firstOperation flag
 *   being set at certain points.
 *   ** should be corrected now **
 *
 *   2+3*4*5+6 isn't correct.  Extra "push eax" in there.
 *   ** also corrected **
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include "compile.h"
#include "output.h"
#include "token.h"
#include "list.h"

List stack;
int firstOperation;
int compStackSize = 0;

// NOTE: here's an idea:
/*
operationIndex *opIndex = {
{ TNumericLiteral, pushToStack }
{ TIdentifierLiteral, pushToStack }
{ TCharacterLiteral, pushToStack }
{ TFloatLiteral, pushToStack }
{ TAdd, performOperation }
};
*/

int compileLine(List *srcList) {
  struct Node *node;
  Token *tok;
  Token *test;

  firstOperation = 1;
  compStackSize  = 0;

  listClear(&stack);


  node = (struct Node *)srcList->first;
  while(/*src->next != NULL*/ node != NULL) {
    tok = node->data;
    //printf("_token(%d)(%s)(%d)\n", tok->type, node->data->name, node->next);
    node = (struct Node *)node->next;

    // NOTE: this was a little stupid of me... out of habit, I simply assume, in my
    // head, that 'node' points to the list-node corresponding to, and holding, 'tok'
    // If you look above, this is not that case.  'tok' refers to the current token,
    // while 'node' is advanced to the next token.  Keep this in mind when coding!
    // I try and avoid these pitfalls when rewriting the compiler with inCode.

    // little kludge-fix, so that popStack wont pop two variables near the end
    // of an equation, when not needed -> this = (2 + 2) * 3;
    //if(node->data->type == TEqual) firstOperation = 0;

    switch(tok->type) {
      case TNumericLiteral:
      case TIdentifier:
//  	  case TStringLiteral:
      case TCharacterLiteral:
		  case TFloatLiteral:
        //stackTok = node->data;
        // have to do this because listAdd will change src->next to NULL!
        //node = (struct Node *)node->next;

        // add this token to our stack
        //printf("test\n");
        listAdd(&stack, tok);
        break;
      case TMul:
        popStack();
        perform1RegOperation("mul");

        // NOTE: Might need to pass node->data->type _AND_ node->next->data->type
        // If there are two numbers in a row, next in the token line, then
        // a push to the stack is required (because both eax and ebx will
        // be overwritten, and eax is the current running total).
        pushResult(node->data->type, node->next->data->type);
        break;
      case TDiv:
        popStack();
        perform1RegOperation("div");
        pushResult(node->data->type, node->next->data->type);
        // div's gonna have to know register/memory operand sizes...
        // might have to xor edx... *NOTE*
        break;
      case TAdd:
        popStack();
        perform2RegOperation("add");
        pushResult(node->data->type, node->next->data->type);
        break;
      case TSub:
        popStack();
        perform2RegOperation("sub");
        pushResult(node->data->type, node->next->data->type);
        break;
      case TEqual:

        /* If the stack isn't empty yet, that means the equation we've
         * been working on is a simple one (ie, y = 5) with only one
         * operand.  Originally I wanted to special case this and move
         * the operand into memory, however, other parts of this code
         * rely on eax being updated, and it would also break single
         * varible expressions (ie, 5) in statements such as while(n).
         *
         * So, instead, this is the perfect candidate for a peek hole
         * optimizer.  "mov eax, x; mov [mem], eax" -> "mov [mem], x"
         */
        if(stack.first != NULL) {
          test = listTop(&stack);
          printf("-> %li %s\n", test->type, test->name);
          outputString("\tmov eax, ");
          //outputString(test->name);
          outputTokenName(test);
          outputString("\n");
          listRemove(&stack, test);
        }
        outputString("\tmov [");
        outputString(node->data->name);
        outputString("], eax\n");

        // since the TEqual section uses TEqual, and the
        // next identifier, advance (otherwise the identifier
        // used above will be pushed to the stack at the end)
        node = node->next;
        break;
      //case TCMP0:
      //  outputString("\tcmp eax, 0\n"); // set flags (needed?)
      //  break;
      case TJNE:
        //outputString("\tcmp eax, 0\n"); // set flags (needed?)
        // gcc only does this with single variable conditions,
        // eg:  if(x) ...
        outputString("\tjne ");
        outputString(tok->name);
        outputString("\n");
        break;
      case TJE:
        outputString("\tje ");
        outputString(tok->name);
        outputString("\n");
        break;
      case TJMP:
        outputString("\tjmp ");
        outputString(tok->name);
        outputString("\n");
        break;
      case TLabel:
        outputString("  ");
        outputString(tok->name);
        outputString(":\n");
        break;
      default:
        outputString("Default\n");
        break;
    }

  }
  return 0;
}

int popStack(void) {
  Token *operand = listTop(&stack);

  // if there's nothing in our maintained stack, then we must get the
  // result from the computers actual stack
  // NOTE: is this right!?  ebx?  Will this ever happen?
  if(stack.first == NULL) {
    outputString("\tpop ebx\n");
    compStackSize--;
    return 0;
  }
  // else?
  // pop into ebx
  outputString("\tmov ebx, ");
  //outputString(operand->name);
  outputTokenName(operand);
  outputString("\n");
  listRemove(&stack, operand);

  if(firstOperation) {
    firstOperation = 0;

    /* NOTE: this isn't 100% right... at the end of the equation (before =), the stack
     * will be exhausted, and therefore the bottom of this function will set firstOperation
     * to 1, and this function will think it has to load two operands... not so!  It's
     * the end of the equation, it simply needs one... see below...
     *
     * -> Should work okay with the compStackSize fix
     */
    if(stack.first == NULL) {
      if(compStackSize != 0) {
        outputString("\tpop eax\n");
        compStackSize--;
      }
    }
    else {
      // pop into (e)ax
      operand = listTop(&stack);
      outputString("\tmov eax, ");
      /*
      if(operand->type == TIdentifier) outputString("[");
      outputString(operand->name);
      if(operand->type == TIdentifier) outputString("]");
      */
      outputTokenName(operand);
      outputString("\n");

      listRemove(&stack, operand);
    }
  }

  // ...
  // If we've exhausted the stack, then pretend like we're at
  // the first operation again (meaning we need two operands,
  // in eax, and ebx, not just the one in ebx)
  // NOTE: UNLESS the next token is '=', then only one token is sufficient!
  if(stack.first == NULL) {
    firstOperation = 2;
  }

  //printf("popStack()\n");
  return 0;
}

/**
 * Checks if a stack push is needed
 * Based on the next two token types, this routine will
 * output a "push" to the stack, if required.
 * @return 0 if okay, !0 on error
 */
int pushResult(int nextOpType, int afterNextOpType) {
  if(nextOpType == TNumericLiteral   ||
     nextOpType == TIdentifier       ||
     // nextOpType == TStringLiteral    ||
     nextOpType == TCharacterLiteral ||
		 nextOpType == TFloatLiteral) {
		
    if(afterNextOpType == TNumericLiteral   ||
       afterNextOpType == TIdentifier       ||
       // afterNextOpType == TStringLiteral    ||
       afterNextOpType == TCharacterLiteral ||
	  	 afterNextOpType == TFloatLiteral) {

        outputString("\tpush eax\n");
        compStackSize++;

     }
  }

  return 0;
}

int perform2RegOperation(char *opcode) {
  //char line[50];

  // output opcode eax, ebx
  //printf("performOperation(%s)\n", opcode);
  outputString("\t");
  outputString(opcode);
  outputString(" eax, ebx\n");
  return 0;
}

int perform1RegOperation(char *opcode) {
  outputString("\t");
  outputString(opcode);
  outputString(" ebx\n");
  return 0;
}