/*
 * InCode - [ list.c ]
 * (c) 2002, J. Weeks
 * jweeks@subdimension.com
 *
 * The following is some code used for creating and maintaining a linked
 * list of tokens.  It isn't overly general purpose... it's used in a few
 * different places in this project, but probably couldn't be used outside
 * of a compiler without changes.
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "token.h"
#include "list.h"

int listCreate(List *list) {
  list->first = NULL;
  list->numItems = 0;

  return 0;
}

struct Node *listCreateNode(Token *tok) {
  struct Node *newNode = (struct Node *)malloc(sizeof(struct Node));
  newNode->next = NULL;
  newNode->data = tok;

  return newNode;
}

int listAdd(List *list, Token *tok) {
  struct Node *newNode = listCreateNode(tok);
  struct Node *search;

  if(list->first == NULL)
    list->first = newNode;
  else {
    search = list->first;
    // traverse the linked list to find a space for this token
    while(search->next != NULL) search = (struct Node *)search->next;

    // put our token in the next available slot
    search->next = (struct Node *)newNode;
  }

  list->numItems++;
  //printf("tokenAdd (%s)\n", tok->name);
  return 0;
}

int listClear(List *list) {
  struct Node *node;
  struct Node *prev;

  // deallocate any memory used (remember, "List" is not allocated,
  // so do not try to deallocate it!)
  node = (struct Node *)list->first;
  while(node != NULL) {
    prev = node;
    node = (struct Node *)node->next;
    free(prev->data);
    free(prev);
  }
  list->first = NULL;
  list->numItems = 0;
  return 0;
}

Token *listTop(List *list) {
  struct Node *node = (struct Node *)list->first;

  if(node == NULL) return (Token *)-1;

  // traverse the linked list to find the end of the list
  while(node->next != NULL) node = (struct Node *)node->next;

  return node->data;
}

// NOTE: currently doesn't work if "tok" isn't in the list!!!
int listRemove(List *list, Token *tok) {
  struct Node *node = (struct Node *)list->first;
  //struct Node *skip;

  // remove the first token
  if(node->data == (Token *)tok) {
    list->first = node->next;
    return 0;
  }

  // traverse the linked list to find the end of the list
  while(node->next->data != (Token *)tok) node = (struct Node *)node->next;

  //skip = (Token *)tokList->next;
  //skip = (Token *)skip->next;
  //tokList->next = (struct Token *)skip;
  node->next = (struct Node *)node->next->next;
  list->numItems--;
  return 0;
}

int listPrintAll(List *list) {
  struct Node *node = (struct Node *)list->first;

  //if(node == NULL) return -1;

  // traverse the list, printing out all tokens along the way
  while(node != NULL) {
    tokenPrint(node->data);
    node = (struct Node *)node->next;
  }
  return 0;
}
