/*
 * InCode - [ parse.c ]
 * (c) 2002, J. Weeks
 * jweeks@subdimension.com
 *
 * The purpose of this module is to take inCode language input, veryify it,
 * and format it into an easily readable order that the compiler can
 * read incredibly quickly and not have to worry about possible errors.
 *
 * eg: x = 1 + 2 * 3;         // translates to:
 *   1, 2, 3, *, +, =, x      // meaning:
 *                            // move 1 into eax
 *                            // push eax to stack, move 2 to eax
 *                            // push eax to stack, move 3 to eax
 *                            // pop stack to ebx, multiply eax, ebx
 *                            // pop stack to ebx, add eax, ebx
 *                            // store result in x
 *
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>

#include "input.h"
#include "token.h"
#include "error.h"
#include "parse.h"
#include "list.h"

List parseTokens;

// the following is just some debugging shit I came up with
// when this thing wasn't working properly.  As always, it
// was one line of offending code, throwing the whole thing
// off!!!
#define spaces \
  for(debug = 0; debug <= debugTot; debug++) printf(" ")

#define spUp   debugTot += 2
#define spDown debugTot -= 2

int debug;
int debugTot = 0;

int currentLabelNum = 0;

// the look ahead token, will always contain the next token to
// be read
Token *lookTok;

// define some new tokens used strictly in the last step (compileLine)...
// --
// negateTok, when read by the compiler, will "neg" the current sub-total
// which will (always?) be in eax.
//Token negateTok = { NULL, "neg", TNegate, 0, 0.0, 0 };
// negateTok cannot be a static token, because of the 'next' pointer
// element.

int parseLine(void) {

  //spUp; spaces;
  //printf("+parseLine()\n");

  // clear the last line of tokens
  listClear(&parseTokens);

  // prime the token getting engine (getting the look ahead token in the process)
  // ruines any lookahead token that might be there (if... else)
  //parseGet(TAny);

  parseStatement();
  //parseGet(TEOL);
  //NOTE: should be at end of parseStatement? Don't think so... not all statements
  // end in a ; if(statement) for example.
  //if(lookTok->type != TEOL) errExpected("semi-colon");

  //spaces;
  //printf("-parseLine()\n");
  //spDown;
  return 0;
}

//int parseLookAhead(void) {
//  lookTok = tokenGet();
//  return 0;
//}

int parseBlock(void) {
  if(lookTok->type == TLeftBrace) {
    parseGet(TLeftBrace);

    while(lookTok->type != TRightBrace) {
      parseStatement();
      //parseGet(TEOL);
    }

    parseGet(TRightBrace);

  } else parseStatement();
  return 0;
}

int parseStatement(void) {
  switch(lookTok->type) {
    case TIf:
      parseIf();
      break;
    case TFor:
      parseFor();
      break;
    case TWhile:
    case TUntil:
      parsePreTestLoop();
      break;
    case TDo:
      parsePostTestLoop();
      break;
    /*
    case TLoop:
      parseNLoop();
      break;
    */
    default:
      // start reading the line (assignment expression)
      parseAssignment();
      // not all statements have a semi-colon (ie, if(statement) {} )
      if(lookTok->type == TEOL) parseGet(TEOL);
      break;
  }
  return 0;
}

/**
 * Parse an assignment statement.
 * Parse a statement in the form of "x = expression"
 * @return 0 on success
 */
int parseAssignment(void) {
  Token *ident;
  Token *equal;

//  spUp; spaces;
//  printf("+parseAssignment()\n");
  ident = parseGet(TIdentifier);
  equal = parseGet(TEqual);

  parseExpression();

  listAdd(&parseTokens, equal);
  listAdd(&parseTokens, ident);

//  spaces;
//  printf("-parseAssignment()\n");
//  spDown;
  return 0;
}

/** parse an if statement
 * if (statement) block else block;
 * eg.
 *    if (this == 2) that = 3;
 *    if (this == 2) {
 *      stuff = 4; something.callMe();
 *    } else blop = null;
 */
int parseIf(void) {
  Token *labelEndIf;
  Token *labelElse;
  Token *JMPEndIf;
  Token *JMPElse;

  parseGet(TIf);
  parseGet(TLeftBracket);

  parseStatement();

  parseGet(TRightBracket);

  // add branch token (jump if not true, to else
  // (or end of if, when no else exists)
  JMPElse = tokenNewType(TJNE);
  newLabel(JMPElse->name);
  //strcpy(labelEndIf->name, JMPElse->name);

  labelElse = tokenNewType(TLabel);
  strcpy(labelElse->name, JMPElse->name);

  listAdd(&parseTokens, JMPElse);

  parseBlock();
  //listAdd(&parseTokens, labelElse);
  //parseGet(TEOL);	// temp. fix for now

  if(lookTok->type == TElse) {
    parseGet(TElse);

    labelEndIf = tokenNewType(TLabel);
    newLabel(labelEndIf->name);
    JMPEndIf = tokenNewType(TJMP);
    strcpy(JMPEndIf->name, labelEndIf->name);

    listAdd(&parseTokens, JMPEndIf);

    listAdd(&parseTokens, labelElse);
    parseBlock();

    listAdd(&parseTokens, labelEndIf);
  }
  else listAdd(&parseTokens, labelElse);
  return 0;
}

int parsePreTestLoop(void) {
  Token *labelStart;
  Token *labelEnd;
  Token *conditionalJump;
  Token *jump;
  Token *loopType;

  // will be either while or until
  loopType = parseGet(TAny);

  labelStart = tokenNewType(TLabel);
  labelEnd   = tokenNewType(TLabel);
  newLabel(labelStart->name);
  newLabel(labelEnd->name);

  listAdd(&parseTokens, labelStart);

  parseGet(TLeftBracket);
  parseStatement();
  parseGet(TRightBracket);

  if(loopType->type == TWhile)
    conditionalJump = tokenNewType(TJNE);
  else
    conditionalJump = tokenNewType(TJE);

  strcpy(conditionalJump->name, labelEnd->name);
  listAdd(&parseTokens, conditionalJump);

  jump = tokenNewType(TJMP);
  strcpy(jump->name, labelStart->name);

  parseBlock();
  listAdd(&parseTokens, jump);
  listAdd(&parseTokens, labelEnd);
  return 0;
}

int parsePostTestLoop(void) {
  Token *labelStart;
  Token *labelEnd;
  Token *conditionalJump;
  Token *jump;
  Token *loopType;

  parseGet(TDo);

  labelStart = tokenNewType(TLabel);
  labelEnd   = tokenNewType(TLabel);
  newLabel(labelStart->name);
  newLabel(labelEnd->name);

  listAdd(&parseTokens, labelStart);

  parseBlock();

  // will be either while or until
  loopType = parseGet(TAny);

  if(loopType->type == TWhile)
    conditionalJump = tokenNewType(TJNE);
  else
    conditionalJump = tokenNewType(TJE);

  //---
  parseGet(TLeftBracket);
  parseStatement();
  parseGet(TRightBracket);

  strcpy(conditionalJump->name, labelEnd->name);
  listAdd(&parseTokens, conditionalJump);

  jump = tokenNewType(TJMP);
  strcpy(jump->name, labelStart->name);

  listAdd(&parseTokens, jump);
  listAdd(&parseTokens, labelEnd);

  parseGet(TEOL);
  return 0;
}


int parseFor(void) {

  parseGet(TFor);
  parseGet(TLeftBracket);

  parseStatement();
  parseStatement();
  parseStatement();

  parseGet(TRightBracket);

  parseBlock();
  return 0;
}

/*
int parseNLoop(void) {
  Token *labelStart;
  Token *labelEnd;
  Token *moveToCX;

  parseGet(TLoop);
  parseGet(TLeftBracket);

  parseStatement();

  parseGet(TRightBracket);

  parseBlock();

  // problem: I intended to use the loop operand to make this
  // simple loop construct... but loop can only jump to an 8-bit
  // relative address.  Not too useful!
}
*/

/**
 * Parse an expression.
 * An expression is any combination of term +- term.
 * @return 0 on success
 */
int parseExpression(void) {
  Token *tok;

//  spUp; spaces;
//  printf("+parseExpression()\n");
  parseTerm();

  tok = lookTok;
  while(tok->type == TAdd || tok->type == TSub) {
    switch(tok->type) {
      case TAdd: parseAdd(); break;
      case TSub: parseSub(); break;
    }
    tok = lookTok;
  }

//  spaces;
//  printf("-parseExpression()\n");
//  spDown;
  return 0;
}

/**
 * Parse a term.
 * A term is an expression in the form xy or x/y.
 * @return 0 on success.       tok = lookTok;

 */
int parseTerm(void) {
  Token *tok;

//  spUp; spaces;
//  printf("+parseTerm()\n");
  parseFactor();

  tok = lookTok;
  while(tok->type == TMul || tok->type == TDiv) {
    switch(tok->type) {
      case TMul: parseMul(); break;
      case TDiv: parseDiv(); break;
    }
    tok = lookTok;
  }

//  spaces;
//  printf("-parseTerm()\n");
//  spDown;
  return 0;
}


/**
 * Parse a factor.
 * A factor is a single individual item (a numeric,
 * or string literal, an identifier, or the
 * result of another expression).
 * @return 0 on success.
 */
int parseFactor(void) {
  Token *tok;

//  spUp; spaces;
//  printf("+parseFactor()\n");

  tok = lookTok;

  if(tok->type == TLeftBracket) {
    parseGet(TLeftBracket);
    parseExpression();
    parseGet(TRightBracket);
  }

  else if(tok->type == TIdentifier)
    parseIdentifier();
  // the next two if's may not work properly.
  // unary plus, ignore it, and continue parsing the factor
  else if(tok->type == TAdd) {
    printf("factor: unary +\n");
    parseGet(TAdd);
    parseFactor();
  }
  // unary minus, if it's a constant, negate it here, else negate it
  // in code.
  else if(tok->type == TSub) {
    printf("factor, unary -\n");
    parseGet(TSub);
    if(lookTok->type == TNumericLiteral) {
      lookTok->value = -lookTok->value;
      parseFactor();
    } else {
      parseFactor();
      listAdd(&parseTokens, tokenNewType(TNegate));
      // a TNegate token means to negate the current sub-total (?) in eax (?)
    }
  }
  else {
    // NOTE: possible way to handle +- unary ops,
    // may not handle -(x/y), the above should handle this (?)
    // if(lookTok->type = TSub) negate = true; else negate = false;
    // if(lookTok->type = TAdd) parseGet(TAdd); // unary + does nothing, so we do nothing!
    tok = parseGet(TNumericLiteral);
    listAdd(&parseTokens, tok);
    //if(negate) tokenAdd(negateTok);
  }

//  spaces;
//  printf("-parseFactor()\n");
//  spDown;
  return 0;
}

/**
 * Parse an identifier.
 * Parses a variable of method identifier.
 * @return 0 on success.
 */
int parseIdentifier(void) {
  Token *tok;

//  spUp; spaces;
//  printf("+parseIdentifier()\n");

  tok = parseGet(TIdentifier);

  if(lookTok->type == TLeftBracket) {
    parseGet(TLeftBracket);
    parseGet(TRightBracket);
    // NOTE: call token->name()
    listAdd(&parseTokens, tok);

  } else {
    // load identifier                   tokenAdd(divide);
    listAdd(&parseTokens, tok);
  }

//  spaces;
//  printf("-parseIdentifier()\n");
//  spDown;
  return 0;
}

/**
 * Parse an addition statement.
 * @return 0 on success.
 */
int parseAdd(void) {
  Token *plus;

//  spUp; spaces;
//  printf("+parseAdd()\n");
  plus = parseGet(TAdd);

  /* Here's a possible way to handle unary +- for the add routine,
   *     if x + +y, reduce to x + y
   * and if x + -y, reduce to x - y
   * Problems arise in assignments like x = -1, however.
   * And methods like function(x, +5, -1);
   * This is probably better handled in factor.
   * Actually... we can perform this optomization here, _AND_ take
   * care of it in factor.  In this case, our optimization will
   * occur before parseFactor, and it'll never know!
   * Saves unneeded code, eg. y = 5 + -x
   *   5 add neg x -> 4 sub x
   */
  if(lookTok->type == TAdd || lookTok->type == TSub)
    plus = parseGet(TAny);

  parseTerm();
  listAdd(&parseTokens, plus);

//  spaces;
//  printf("-parseAdd()\n");
//  spDown;
  return 0;
}

/**
 * Parse a subtraction statement.
 * @return 0 on success.
 */
int parseSub(void) {
  Token *minus;

  minus = parseGet(TSub);

  // Pretty straightforward, x - +1 = x - 1
  if(lookTok->type == TAdd)
    parseGet(TAdd);

  // and x - -y = x + y
  if(lookTok->type == TSub) {
    parseGet(TSub);
    minus->type = TAdd;
    minus->name[0] = '+';	// NOTE: can be deleted... just for my own reference
  }

  parseTerm();
  listAdd(&parseTokens, minus);

  return 0;
}

/**
 * Parse a multiplication statement.
 * @return 0 on success.
 */
int parseMul(void) {
  Token *multiply;
  //char negate = 0;

  multiply = parseGet(TMul);

  // I don't know if the following code is worth the effort...
  // it doesn't make a difference like the code for parseAdd and
  // parseSub do.  In fact, in the simple trivial case,
  // y = 2 * -3, for example, it produces _less_ efficient code
  // (using a negate token, as apposed to simply negating 3
  // at compile time).
  /*
  // x * +y = x * y
  if(lookTok->type == TAdd)
    parseGet(TAdd);		// forget about the unary +

  // x * -y = -(x * y)
  if(lookTok->type == TSub) {
    parseGet(TSub);		// get the unary -, and set our negate flag
    negate = 1;
  }
  */

  parseFactor();
  listAdd(&parseTokens, multiply);

  //if(negate) tokenAdd(tokenNewType(TNegate));

  return 0;
}

/**
 * Parse a division statement.
 * @return 0 on success.
 */
int parseDiv(void) {
  Token *divide;
  //char negate = 0;

  divide = parseGet(TDiv);

  /*
  // x / +y = x / y
  if(lookTok->type == TAdd)
    parseGet(TAdd);

  // x / -y = -(x / y)
  if(lookTok->type == TSub) {
    parseGet(TSub);
    negate = 1;
  }
  */

  parseFactor();
  listAdd(&parseTokens, divide);

  //if(negate) tokenAdd(tokenNewType(TNegate));

  return 0;
}

/**
 * Get a token
 * Returns a token of type 'type'  Actually returns the look ahead
 * token, and "re-gets" the look ahead.
 * @param type The type of token to return, -1 if it doesn't matter.
 * @return the token.
 */
Token *parseGet(long type) {
  Token *ret = lookTok;
  char errStr[20];

  if (type != TAny && ret->type != type) {
    sprintf(errStr, "found token(%li), token(%li)", ret->type, type);
    errExpected(errStr);
  }
  lookTok = tokenGet();

  return ret;
}

int newLabel(char *labelName) {
  // NOTE: might need to change label size here
  //char *labelName = (char *)malloc(10);

  sprintf(labelName, "lbl_%d", currentLabelNum++);
  return 0;
}