/*
 * InCode - [ token.c ]
 * (c) 2002, J. Weeks
 * jweeks@subdimension.com
 *
 * This is the inCode tokenizer engine.  Given an input stream (from the input modules)
 * it will break it up into tokens, which will then be sent to the compiler.
 *
 * I haven't yet decided wether to implement optimization into this section or not.
 *   - optomization else where would probably require another stream of tokens, thus
 *     more memory overhead
 *   - optomization else where could be more easily turn on/off with a compiler
 *     flag.
 * Possible memory optimization
 *   - only allocate the name string (32 bytes!) for those that actually need it!
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "token.h"
#include "input.h"
#include "iswhat.h"
#include "error.h"

//Token token;

TokenLookupItem tokLookup[] = {
  { TByte,           "byte"  },
  { TWord,           "word"  },
  { TDWord,          "dword" },
  { TQWord,          "qword" },
  { TSingle,         "single"},
  { TDouble,         "double"},
//  { TDWord,           "long" },
//  { TQWord,					 "huge" },
//  { TUnsigned,       "unsigned" },
  { TClass, 				 "class" },
  { TEOL,            ";" },
  { TEnd,						 ";" }, // must be placed _after_ TEOL for parser to work
  { TLeftBracket,    "(" },
  { TRightBracket,   ")" },
  { TLeftBrace,      "{" },
  { TRightBrace,     "}" },
  { TEqual,          "=" },
  { TAdd,            "+" },
  { TSub,            "-" },
  { TDiv,            "/" },
  { TMul,            "*" },
  { TExp,            "^" },
  { TCompEqual,      "==" },		// you'll notice you can use both == or 'is' to compare
  { TCompEqual,			 "is" },		// things
  { TAddEqual,       "+=" },
  { TSubEqual,       "-=" },
  { TDivEqual,       "/=" },
  { TMulEqual,       "*=" },
  { TLeftSqBracket,  "[" },
  { TRightSqBracket, "]" },
  { TIf,						 "if"   },
  { TElse,           "else" },
  { TDo,             "do"   },
  { TWhile,          "while"},
  { TUntil,          "until"},
  { TLoop,           "loop" }, //loop(n) stuff;; // loop n number of times
  { TFor,            "for"  },
  { TBreak,          "break"},
  { TEndList,        "!!!" }
};

/**
 * Initialize Token System.
 * Initializes the token linked list, and any other setup procedures.
 * @return 0 if okay, !0 on error.
 */
int tokenInit(void) {
//  token.next = NULL;
//  token.type = 0;
  return 0;
}

/**
 * Close the Token sytem.
 * Removes any memory occupied by the token system, and releases
 * any resources used.
 * @return 0 if okay, !0 on error.
 */
int tokenClose(void) {
//  tokenClear();

  return 0;
}

/*
int tokenClear(void) {
  Token *tok;
  Token *prev;

  // deallocate any memory used (remember, "token" is not allocated,
  // so do not try to deallocate it!)
  tok = (Token *)token.next;
  while(tok != NULL) {
    prev = tok;
    tok = (Token *)tok->next;
    free(prev);
  }
  token.next = NULL;
  return 0;
}
*/

// useless function now... handled in parseLine
/*
int tokenize(void) {
  //Token *curTok;

*/
  /*
  Token *curTok;
  printf("\n+ tokenize()\n");
  while(tokenGet(curTok) == 0) {
    printf("Token (%c)\n", curTok->name[0]);
    if(curTok->name[0] == 'Q') break;
    tokenDelete(curTok);
  }
  printf("- tokenize()\n");
  return 0;
  */


  /* this function should possibly be redesigned to read in a
   * token at a time, and parse the line for validity, adding to
   * the linked list of tokens.
   *
   * If so, perhaps it should be renamed to parseLine, and moved
   * to a parse.c file
   */
   /*
  printf("+tokenize()\n");
  tokenClear();
  inputLookAhead();

  while( tokenGet() == 0 ) { }
  printf("-tokenize()\n");
  return 0;
}
*/

Token *tokenGet(void) {
  char c;
  Token *tok;

  //printf("+ getToken()\n");
  // create a new token for our data
  tok = tokenNew();

  // read blanks, if there are any
  while ( isWhite(lookAhead) ) inputLookAhead();

  // interpret the incomming data based on the look ahead character

  // alphabetic -> read identifier
  if(isAlpha(lookAhead)) {
    inputIdent(tok->name);
    // check against known identifiers (ie, reserved words)
    tok->type = tokenTypeFromString(tok->name);
    // if not found, just call it a standard identifier
    if(tok->type == -1) tok->type = TIdentifier;
  }
  // numeric -> read number literal
  else if(isNum(lookAhead)) {
    // std format for number base_number (ie, 16_FFEF01EF + 8_0776667)
    inputNumber(tok->name);
    tok->type = TNumericLiteral;
    tok->value = toNumber(tok->name);
  }
  // quote -> read string literal
  else if(isQuote(lookAhead)) {
    switch(lookAhead) {
      case '"':
        tok->type = TStringLiteral;
        break;
      case '\'':
        tok->type = TCharacterLiteral;
        break;
    }
    // NOTE: this will have to change... it reads in tok->name,
    // which will only support 32 character strings.
    inputUntilChar(tok->name, lookAhead);
  }
  // other -> read symbol
  else if(isSymbol(lookAhead)) {
    inputSymbol(tok->name);
    tok->type = tokenTypeFromString(tok->name);
    if(tok->type == -1) {
      printf("unexpected symbol: %s\n", tok->name);
      errExpected("token->symbol");
    }
    // input look ahead?!
  }
  else {
    // I don't this can possibly be executed as long as the above
    // if statement exists
    printf("unexpected token: %c\n", c);
    printf("NOTE: you should _NEVER_ see this!\n");
    errAbort("fix this! :)");
  }

  //tokenAdd(tok);

  //printf("- getToken()\n");
  //if(tok->type == TEOL) return 1;
  //else return 0;
  return tok;
}

Token *tokenNew(void) {
  Token *tokRet;

  tokRet = (Token *)malloc(sizeof(Token));
  //tokRet->next = NULL;
  tokRet->name[0] = '\0';
  tokRet->type = -1;
  tokRet->value = 0;
  tokRet->fvalue = 0.0;

  return tokRet;
}

Token *tokenNewType(long type) {
  Token *newTok = tokenNew();

  newTok->type = type;

  return newTok;
}

int tokenDelete(Token *delTok) {
  printf("+ deltok\n");

  free(delTok);
  return 0;
}

int tokenPrint(Token *tok) {
  printf("name:\t %s\n", tok->name);
  printf("type:\t %li \n", tok->type);
  printf("value:\t %li \n", tok->value);
  printf("fvalue:\t %f \n", tok->fvalue);
  printf("size:\t %li \n", tok->size);
  printf("\n");
  return 0;
}

/*
int tokenPrintAll(void) {
  Token *tokList = (Token *)&token;

  // traverse the list, printing out all tokens along the way
  while(tokList->next != NULL) {
    tokList = (Token *)tokList->next;
    tokenPrint(tokList);
  }
  return 0;
}
*/

/*
int tokenAdd(Token *tok) {
  Token *tokList = (Token *)&token;

  // traverse the linked list to find a space for this token
  while(tokList->next != NULL) tokList = (Token *)tokList->next;

  // put our token in the next available slot
  tokList->next = (struct Token *)tok;
  printf("tokenAdd (%s)\n", tok->name);
  return 0;
}
*/

int tokenTypeFromString(char *str) {
  int i = 0;

  // search through all symbols and, if found, set the type variable
  while( tokLookup[i].type != TEndList) {
    if(strcmp(str, tokLookup[i].looksLike) == 0)
      return tokLookup[i].type;
    i++;
  }

  return -1;
}